# main.py | main Flask routes and error handling
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import flask
from sentry_sdk import capture_exception

import sciolyid.config as config
from sciolyid.web.blueprints import upload, user, verify, about
from sciolyid.web.config import app, logger

app.register_blueprint(about.bp)
app.register_blueprint(user.bp)
if not config.options["disable_upload"]:
    app.register_blueprint(upload.bp)
if not config.options["disable_validation"]:
    app.register_blueprint(verify.bp)


@app.route("/")
def api_index():
    logger.info("index page accessed")
    return "<h1>Hello!</h1><p>This is the index page for the sciolyid internal API.<p>"


@app.errorhandler(400)
def client_error(e):
    capture_exception(e)
    return flask.jsonify(error=str(e)), 400


@app.errorhandler(403)
def not_allowed(e):
    capture_exception(e)
    return flask.jsonify(error=str(e)), 403


@app.errorhandler(404)
def not_found(e):
    return flask.jsonify(error=str(e)), 404


@app.errorhandler(406)
def input_error(e):
    capture_exception(e)
    return flask.jsonify(error=str(e)), 406


@app.errorhandler(500)
def other_internal_error(e):
    capture_exception(e)
    return flask.jsonify(error=str(e)), 500


@app.errorhandler(503)
def internal_error(e):
    capture_exception(e)
    return flask.jsonify(error=str(e)), 503



# util.py | assorted utility functions that are standalone
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import difflib
import functools
import math
import random
from io import BytesIO
from typing import Iterable, Optional

import discord
from PIL import Image


def cache(func=None):
    """Cache decorator based on functools.lru_cache.
    This does not have a max_size and does not evict items.
    In addition, results are only cached by the first provided argument.
    """

    def wrapper(func):
        sentinel = object()

        cache_ = {}
        hits = misses = 0
        cache_get = cache_.get
        cache_len = cache_.__len__

        def _evict():
            """Evicts a random item from the local cache."""
            if len(cache_) > 0:
                cache_.pop(random.choice(tuple(cache_)), 0)

        async def wrapped(*args, **kwds):
            # Simple caching without ordering or size limit
            nonlocal hits, misses
            key = hash(args[0])
            result = cache_get(key, sentinel)
            if result is not sentinel:
                hits += 1
                return result
            misses += 1
            result = await func(*args, **kwds)
            cache_[key] = result
            return result

        def cache_info():
            """Report cache statistics"""
            return functools._CacheInfo(hits, misses, None, cache_len())

        wrapped.cache_info = cache_info
        wrapped.evict = _evict
        return functools.update_wrapper(wrapped, func)

    if func:
        return wrapper(func)
    return wrapper


def black_and_white(input_image_path) -> BytesIO:
    """Returns a black and white version of an image.

    Output type is a file object (BytesIO).

    `input_image_path` - path to image (string) or file object
    """
    with Image.open(input_image_path) as color_image:
        bw = color_image.convert("L")
        final_buffer = BytesIO()
        bw.save(final_buffer, "png")
    final_buffer.seek(0)
    return final_buffer


async def fetch_get_user(user_id: int, ctx=None, bot=None, member: bool = False):
    if (ctx is None and bot is None) or (ctx is not None and bot is not None):
        raise ValueError("Only one of ctx or bot must be passed")
    if ctx:
        bot = ctx.bot
    elif member:
        raise ValueError("ctx must be passed for member lookup")
    if not member:
        return await _fetch_cached_user(user_id, bot)
    if bot.intents.members:
        return ctx.guild.get_member(user_id)
    try:
        return await ctx.guild.fetch_member(user_id)
    except discord.HTTPException:
        return None


@cache()
async def _fetch_cached_user(user_id: int, bot):
    if bot.intents.members:
        return bot.get_user(user_id)
    try:
        return await bot.fetch_user(user_id)
    except discord.HTTPException:
        return None


def prune_user_cache(count: int = 5):
    """Evicts `count` items from the user cache."""
    for _ in range(count):
        _fetch_cached_user.evict()


def spellcheck_list(
    word_to_check: str, correct_list: Iterable[str], abs_cutoff: Optional[int] = None
):
    for correct_word in correct_list:
        if abs_cutoff is None:
            relative_cutoff = math.floor(len(correct_word) / 3)
        else:
            relative_cutoff = abs_cutoff
        if spellcheck(word_to_check, correct_word, relative_cutoff):
            return True
    return False


def spellcheck(worda: str, wordb: str, cutoff: int = 3) -> bool:
    """Checks if two words are close to each other.
    `worda` (str) - first word to compare
    `wordb` (str) - second word to compare
    `cutoff` (int) - allowed difference amount
    """
    worda = worda.lower().replace("-", " ").replace("'", "")
    wordb = wordb.lower().replace("-", " ").replace("'", "")
    shorterword = min(worda, wordb, key=len)
    if worda != wordb:
        if (
            len(tuple(difflib.Differ().compare(worda, wordb))) - len(shorterword)
            > cutoff
        ):
            return False
    return True


def better_spellcheck(
    word: str, correct: Iterable[str], options: Iterable[str]
) -> bool:
    """Allow lenient spelling unless another answer is closer."""
    matches = difflib.get_close_matches(
        word.lower(), map(str.lower, options), n=1, cutoff=(2 / 3)
    )
    if not matches:
        return False
    if matches[0] in map(str.lower, correct):
        return True
    return False



# functions/user.py | functions relating to users
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import time
from typing import Dict, Union

import requests
from flask import abort, session

import sciolyid.config as config

PROFILE_URL = "https://discord.com/api/users/{id}"
AVATAR_URL = "https://cdn.discordapp.com/avatars/{id}/{avatar}.{ext}"
DEFAULT_AVATAR_URL = "https://cdn.discordapp.com/embed/avatars/{discrim}.png"
INVITE_TO_ID_URL = "https://discord.com/api/invites/{code}"

SESSION_EXPIRE: int = 432000  # time (seconds) before expiring the session


class DiscordBotAuth(requests.auth.AuthBase):
    def __call__(self, request):
        request.headers["Authorization"] = (
            "Bot " + os.environ[config.options["bot_token_env"]]
        )
        return request


def fetch_profile(user_id: Union[int, str]) -> Dict[str, str]:
    url = PROFILE_URL.format(id=user_id)
    resp = requests.get(url, auth=DiscordBotAuth(), timeout=10)
    if resp.status_code != 200:
        abort(404, "Failed to fetch profile")
    json: dict = resp.json()

    profile: Dict[str, str] = dict()
    profile["username"] = f"{json['username']}#{json['discriminator']}"
    if json["avatar"]:
        profile["avatar"] = AVATAR_URL.format(
            id=user_id,
            avatar=json["avatar"],
            ext=("gif" if json["avatar"].startswith("a_") else "png"),
        )
    else:
        profile["avatar"] = DEFAULT_AVATAR_URL.format(
            discrim=int(json["discriminator"]) % 5
        )

    return profile


def fetch_server_id():
    if config.options.get("verification_server_id", None):
        return config.options.get("verification_server_id")
    url = INVITE_TO_ID_URL.format(
        code=config.options["verification_server"].split("/")[-1]
    )
    resp = requests.get(url, auth=DiscordBotAuth(), timeout=10)
    if resp.status_code != 200:
        abort(404, "Failed to fetch id")
    json: dict = resp.json()
    config.options["verification_server_id"] = json["guild"]["id"]
    return json["guild"]["id"]


def get_user_id() -> str:
    date: int = int(session.get("date", 0))
    if (time.time() - date) > SESSION_EXPIRE:
        abort(403, "Your session expired")
    uid: str = session["uid"]
    return uid



# github.py | functions for syncing a GitHub repo
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import concurrent.futures
import os
import logging

import filelock
from git import Repo

import sciolyid.config as config

download_logger = logging.getLogger(config.options["name"] + ".git_downloads")


async def download_github(data, category, item):  # pylint: disable=unused-argument
    download_logger.info("syncing github")
    with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
        loop = asyncio.get_event_loop()
        try:
            os.listdir(config.options["download_dir"])
        except FileNotFoundError:
            download_logger.info("doesn't exist, cloning")
            await loop.run_in_executor(executor, _clone)
            download_logger.info("done cloning")
        else:
            download_logger.info("exists, syncing")
            await loop.run_in_executor(executor, _sync)
            download_logger.info("done syncing")


def _clone():
    lock = filelock.FileLock(config.options["download_dir"].strip("/") + ".lock")
    with lock:
        Repo.clone_from(
            config.options["github_image_repo_url"],
            config.options["download_dir"],
            multi_options=["--depth=1"],
        )


def _sync():
    downloads = Repo(config.options["download_dir"])
    downloads.remote("origin").fetch()
    downloads.head.reset(working_tree=True)



# Based on https://redis.io/topics/quickstart
import subprocess
import tarfile
from urllib.request import urlopen
from shutil import copyfileobj

# from https://stackoverflow.com/a/15035466/7941251
with urlopen("http://download.redis.io/redis-stable.tar.gz") as in_stream, open(
    "redis-stable.tar.gz", "wb"
) as out_file:
    copyfileobj(in_stream, out_file)
with tarfile.open("redis-stable.tar.gz") as tar:
    tar.extractall()
subprocess.run("make", cwd="redis-stable")



import datetime
import string

from sciolyid.data import database, logger


async def channel_setup(ctx):
    """Sets up a new discord channel.

    `ctx` - Discord context object
    """
    logger.info("checking channel setup")
    if not database.exists(f"channel:{ctx.channel.id}"):
        database.hset(
            f"channel:{ctx.channel.id}",
            mapping={"item": "", "answered": 1, "prevJ": 20, "prevI": ""},
        )
        # true = 1, false = 0, prevJ is 20 to define as integer
        logger.info("channel data added")
        await ctx.send("Ok, setup! I'm all ready to use!")

    if database.zscore("score:global", str(ctx.channel.id)) is None:
        database.zadd("score:global", {str(ctx.channel.id): 0})
        logger.info("channel score added")

    if ctx.guild is not None:
        channels = map(lambda x: str(x.id), ctx.guild.text_channels)
        database.sadd(f"channels:{ctx.guild.id}", *channels)


async def user_setup(ctx):
    """Sets up a new discord user for score tracking.

    `ctx` - Discord context object
    """
    logger.info("checking user data")
    if database.zscore("users:global", str(ctx.author.id)) is None:
        database.zadd("users:global", {str(ctx.author.id): 0})
        logger.info("user global added")
        await ctx.send("Welcome <@" + str(ctx.author.id) + ">!")

    date = str(datetime.datetime.now(datetime.timezone.utc).date())
    if database.zscore(f"daily.score:{date}", str(ctx.author.id)) is None:
        database.zadd(f"daily.score:{date}", {str(ctx.author.id): 0})
        logger.info("user daily added")

    # Add streak
    if (database.zscore("streak:global", str(ctx.author.id)) is None) or (
        database.zscore("streak.max:global", str(ctx.author.id)) is None
    ):
        database.zadd("streak:global", {str(ctx.author.id): 0})
        database.zadd("streak.max:global", {str(ctx.author.id): 0})
        logger.info("added streak")

    if ctx.guild is not None:
        if database.exists(f"users.server:{ctx.guild.id}"):
            users = map(
                lambda x: x.decode("utf-8"),
                database.zrange(f"users.server:{ctx.guild.id}", 0, -1),
            )
            database.sadd(f"users.server.id:{ctx.guild.id}", *users)
            database.delete(f"users.server:{ctx.guild.id}")
        database.sadd(f"users.server.id:{ctx.guild.id}", str(ctx.author.id))
        logger.info("synced user to server")


def item_setup(ctx, item: str):
    """Sets up a new item for incorrect tracking.

    `ctx` - Discord context object
    `item` - item to setup
    """
    logger.info("checking item data")
    if database.zscore("incorrect:global", string.capwords(item)) is not None:
        logger.info("item global ok")
    else:
        database.zadd("incorrect:global", {string.capwords(item): 0})
        logger.info("item global added")

    if (
        database.zscore(f"incorrect.user:{ctx.author.id}", string.capwords(item))
        is not None
    ):
        logger.info("incorrect item user ok")
    else:
        database.zadd(f"incorrect.user:{ctx.author.id}", {string.capwords(item): 0})
        logger.info("incorrect item user added")

    if (
        database.zscore(f"correct.user:{ctx.author.id}", string.capwords(item))
        is not None
    ):
        logger.info("correct item user ok")
    else:
        database.zadd(f"correct.user:{ctx.author.id}", {string.capwords(item): 0})
        logger.info("correct item user added")

    date = str(datetime.datetime.now(datetime.timezone.utc).date())
    if database.zscore(f"daily.incorrect:{date}", string.capwords(item)) is not None:
        logger.info("item daily ok")
    else:
        database.zadd(f"daily.incorrect:{date}", {string.capwords(item): 0})
        logger.info("item daily added")

    if database.zscore("frequency.item:global", string.capwords(item)) is not None:
        logger.info("item freq global ok")
    else:
        database.zadd("frequency.item:global", {string.capwords(item): 0})
        logger.info("item freq global added")

    if ctx.guild is not None:
        logger.info("no dm")
        if (
            database.zscore(f"incorrect.server:{ctx.guild.id}", string.capwords(item))
            is not None
        ):
            logger.info("item server ok")
        else:
            database.zadd(
                f"incorrect.server:{ctx.guild.id}", {string.capwords(item): 0}
            )
            logger.info("item server added")
    else:
        logger.info("dm context")

    if database.exists(f"session.data:{ctx.author.id}"):
        logger.info("session in session")
        if (
            database.zscore(f"session.incorrect:{ctx.author.id}", string.capwords(item))
            is not None
        ):
            logger.info("item session ok")
        else:
            database.zadd(
                f"session.incorrect:{ctx.author.id}", {string.capwords(item): 0}
            )
            logger.info("item session added")
    else:
        logger.info("no session")


def session_increment(ctx, item: str, amount: int = 1):
    """Increments the value of a database hash field by `amount`.

    `ctx` - Discord context object\n
    `item` - hash field to increment (see data.py for details,
    possible values include correct, incorrect, total)\n
    `amount` (int) - amount to increment by, usually 1
    """
    if database.exists(f"session.data:{ctx.author.id}"):
        logger.info("session active")
        logger.info(f"incrementing {item} by {amount}")
        value = int(database.hget(f"session.data:{ctx.author.id}", item))
        value += int(amount)
        database.hset(f"session.data:{ctx.author.id}", item, str(value))
    else:
        logger.info("session not active")


def incorrect_increment(ctx, item: str, amount: int = 1):
    """Increments the value of an incorrect item by `amount`.

    `ctx` - Discord context object\n
    `item` - item that was incorrect\n
    `amount` (int) - amount to increment by, usually 1
    """
    logger.info(f"incrementing incorrect {item} by {amount}")
    date = str(datetime.datetime.now(datetime.timezone.utc).date())
    database.zincrby("incorrect:global", amount, string.capwords(item))
    database.zincrby(f"incorrect.user:{ctx.author.id}", amount, string.capwords(item))
    database.zincrby(f"daily.incorrect:{date}", amount, string.capwords(item))
    if ctx.guild is not None:
        logger.info("no dm")
        database.zincrby(
            f"incorrect.server:{ctx.guild.id}", amount, string.capwords(item)
        )
    else:
        logger.info("dm context")
    if database.exists(f"session.data:{ctx.author.id}"):
        logger.info("session in session")
        database.zincrby(
            f"session.incorrect:{ctx.author.id}", amount, string.capwords(item)
        )
    else:
        logger.info("no session")


def score_increment(ctx, amount: int = 1):
    """Increments the score of a user by `amount`.

    `ctx` - Discord context object\n
    `amount` (int) - amount to increment by, usually 1
    """
    logger.info(f"incrementing score by {amount}")
    date = str(datetime.datetime.now(datetime.timezone.utc).date())
    database.zincrby("score:global", amount, str(ctx.channel.id))
    database.zincrby("users:global", amount, str(ctx.author.id))
    database.zincrby(f"daily.score:{date}", amount, str(ctx.author.id))
    if ctx.guild is not None and database.exists(f"race.data:{ctx.channel.id}"):
        logger.info("race in session")
        database.zincrby(f"race.scores:{ctx.channel.id}", amount, str(ctx.author.id))
    else:
        logger.info("dm context")


def streak_increment(ctx, amount: int):
    """Increments the streak of a user by `amount`.

    `ctx` - Discord context object\n
    `amount` (int) - amount to increment by, usually 1.
    If amount is None, the streak is ended.
    """

    if amount is not None:
        # increment streak and update max
        database.zincrby("streak:global", amount, ctx.author.id)
        if database.zscore("streak:global", ctx.author.id) > database.zscore(
            "streak.max:global", ctx.author.id
        ):
            database.zadd(
                "streak.max:global",
                {ctx.author.id: database.zscore("streak:global", ctx.author.id)},
            )
    else:
        database.zadd("streak:global", {ctx.author.id: 0})



# stats.py | commands for bot statistics
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
from io import BytesIO, StringIO
from typing import Literal

import discord
import numpy as np
import pandas as pd
from discord import app_commands
from discord.ext import commands

import sciolyid.config as config
from sciolyid.data import database, logger
from sciolyid.functions import CustomCooldown, send_leaderboard
from sciolyid.util import fetch_get_user


def auto_options(options):
    async def inner(_: discord.Interaction, __: str):
        return [app_commands.Choice(name=name, value=name) for name in options]

    return inner


class Stats(commands.Cog):
    def __init__(self, bot):
        self.bot = bot

    @staticmethod
    def generate_series(database_key):
        """Generates a pandas.Series from a Redis sorted set."""
        logger.info("generating series")
        data = database.zrevrangebyscore(database_key, "+inf", "-inf", withscores=True)
        return pd.Series(
            {e[0]: e[1] for e in map(lambda x: (x[0].decode("utf-8"), int(x[1])), data)}
        )

    @staticmethod
    def generate_dataframe(database_keys, titles):
        """Generates a pandas.DataFrame from multiple Redis sorted sets."""
        pipe = database.pipeline()
        for key in database_keys:
            pipe.zrevrangebyscore(key, "+inf", "-inf", withscores=True)
        result = pipe.execute()
        df = pd.DataFrame()
        for i, item in enumerate(result):
            df.insert(
                len(df.columns),
                titles[i],
                pd.Series(
                    {
                        e[0]: e[1]
                        for e in map(lambda x: (x[0].decode("utf-8"), int(x[1])), item)
                    }
                ),
            )
        df = df.fillna(value=0).astype(int)
        return df

    async def convert_users(self, df):
        """Converts discord user ids in DataFrames or Series indexes to usernames."""
        current_ids = df.index
        new_index = []
        for user_id in current_ids:
            user = await fetch_get_user(int(user_id), bot=self.bot, member=False)
            if user is None:
                new_index.append("User Unavailable")
            else:
                new_index.append(f"{user.name}#{user.discriminator}")
        df.index = new_index
        return df

    # give frequency stats
    @commands.hybrid_command(
        help=f"- Gives info on command/{config.options['id_type']} frequencies",
        aliases=["freq"],
    )
    @commands.check(CustomCooldown(5.0, bucket=commands.BucketType.channel))
    @app_commands.describe(scope="type of frequency", page="page number")
    @app_commands.autocomplete(
        scope=auto_options((
                "command",
                config.options["id_type"],
                "commands",
                "c",
                config.options["id_type"][0],
            )
        )
    )
    async def frequency(
        self,
        ctx: commands.Context,
        scope: str = "",
        page: int = 1,
    ):
        logger.info("command: frequency")

        if scope in ("command", "commands", "c"):
            database_key = "frequency.command:global"
            title = "Most Frequently Used Commands"
        elif scope in (config.options["id_type"], config.options["id_type"][0]):
            database_key = "frequency.item:global"
            title = f"Most Frequent {config.options['id_type'].title()}"
        else:
            await ctx.send(
                f"**Invalid Scope!**\n*Valid Scopes:* `commands, {config.options['id_type']}`",
                ephemeral=True,
            )
            return

        await send_leaderboard(ctx, title, page, database_key)

    # give bot stats
    @commands.hybrid_command(
        help="- Gives statistics on different topics",
        usage="[topic]",
        aliases=["stat"],
    )
    @commands.check(CustomCooldown(5.0, bucket=commands.BucketType.channel))
    @app_commands.describe(topic="stats on what?")
    async def stats(
        self,
        ctx: commands.Context,
        topic: Literal["scores", "usage", "help", "score", "s", "u"] = "help",
    ):
        logger.info("command: stats")

        if topic in ("scores", "score", "s"):
            topic = "scores"
        elif topic in ("usage", "u"):
            topic = "usage"
        elif topic in ("help", ""):
            topic = "help"
        else:
            valid_topics = ("help", "scores", "usage")
            await ctx.send(
                f"**`{topic}` is not a valid topic!**\nValid Topics: `{'`, `'.join(valid_topics)}`",
                ephemeral=True,
            )
            return

        if ctx.interaction is not None:
            await ctx.typing()

        embed = discord.Embed(
            title="Bot Stats",
            type="rich",
            color=discord.Color.blue(),
        )

        if topic == "help":
            embed.description = (
                "**Available statistic topics.**\n"
                + "This command is in progress and more stats may be added. "
                + "If there is a statistic you would like to see here, "
                + "please let us know in the support server."
            )
            embed.add_field(
                name="Scores",
                value=f"`{config.options['prefixes'][0]}stats [scores|score|s]`\n*Displays stats about scores.*",
            ).add_field(
                name="Usage",
                value=f"`{config.options['prefixes'][0]}stats [usage|u]`\n*Displays stats about usage.*",
            )

        elif topic == "scores":
            embed.description = "**Score Statistics**"
            scores = self.generate_series("users:global")
            scores = scores[scores > 0]
            c, d = np.histogram(scores, bins=range(0, 1100, 100), range=(0, 1000))
            c = (c / len(scores) * 100).round(1)
            embed.add_field(
                name="Totals",
                inline=False,
                value="**Sum of top 10 user scores:** `{:,}`\n".format(
                    scores.nlargest(n=10).sum()
                )
                + "**Sum of all positive user scores:** `{:,}`\n".format(scores.sum()),
            ).add_field(
                name="Computations",
                inline=False,
                value="**Mean of all positive user scores:** `{:,.2f}`\n".format(
                    scores.mean()
                )
                + "**Median of all positive user scores:** `{:,.1f}`\n".format(
                    scores.median()
                ),
            ).add_field(
                name="Distributions",
                inline=False,
                value=f"**Number of users with scores over mean:** `{len(scores[scores > scores.mean()])}`\n"
                + "**Percentage of users with scores over mean:** `{:.1%}`".format(
                    len(scores[scores > scores.mean()]) / len(scores)
                )
                + "\n**Percentage of users with scores between:**\n"
                + "".join(
                    f"\u2192 *{d[i]}-{d[i+1]-1}*: `{c[i]}%`\n"  # \u2192 is the "Rightwards Arrow"
                    for i in range(len(c))
                ),
            )

        elif topic == "usage":
            embed.description = "**Usage Statistics**"

            today = datetime.datetime.now(datetime.timezone.utc).date()
            past_month = pd.date_range(  # pylint: disable=no-member
                today - datetime.timedelta(29), today
            ).date
            keys = ("users:global",) + tuple(
                f"daily.score:{str(date)}" for date in past_month
            )
            titles = tuple(
                reversed(range(1, 32))
            )  # label columns by # days ago, today is 1 day ago
            month = self.generate_dataframe(keys, titles)
            total = month.loc[:, 31]
            month = month.loc[:, 30:1]  # remove totals column
            month = month.loc[(month != 0).any(1)]  # remove users with all 0s
            week = month.loc[:, 7:1]  # generate week from month
            week = week.loc[(week != 0).any(1)]
            today = week.loc[:, 1]  # generate today from week
            today = today.loc[today != 0]

            channels_see = len(tuple(self.bot.get_all_channels()))
            channels_used = int(database.zcard("score:global"))

            embed.add_field(
                name="Today (Since midnight UTC)",
                inline=False,
                value="**Accounts that answered at least 1 correctly:** `{:,}`\n".format(
                    len(today)
                )
                + f"**Total {config.options['id_type']} answered correctly:** `{today.sum():,}`\n",
            ).add_field(
                name="Last 7 Days",
                inline=False,
                value="**Accounts that answered at least 1 correctly:** `{:,}`\n".format(
                    len(week)
                )
                + f"**Total {config.options['id_type']} answered correctly:** `{week.sum().sum():,}`\n",
            ).add_field(
                name="Last 30 Days",
                inline=False,
                value="**Accounts that answered at least 1 correctly:** `{:,}`\n".format(
                    len(month)
                )
                + f"**Total {config.options['id_type']} answered correctly:** `{month.sum().sum():,}`\n",
            ).add_field(
                name="Total",
                inline=False,
                value="**Channels the bot can see:** `{:,}`\n".format(channels_see)
                + "**Channels that have used the bot at least once:** `{:,} ({:,.1%})`\n".format(
                    channels_used, channels_used / channels_see
                )
                + "*(Note: Deleted channels or channels that the bot can't see anymore are still counted).*\n"
                + "**Accounts that have used any command at least once:** `{:,}`\n".format(
                    len(total)
                )
                + "**Accounts that answered at least 1 correctly:** `{:,} ({:,.1%})`\n".format(
                    len(total[total > 0]), len(total[total > 0]) / len(total)
                ),
            )

        await ctx.send(embed=embed)
        return

    # export data as csv
    @commands.hybrid_command(help="- Exports bot data as a csv")
    @commands.check(CustomCooldown(60.0, bucket=commands.BucketType.channel))
    async def export(self, ctx: commands.Context):
        logger.info("command: export")

        files = []

        async def _export_helper(database_keys, header, filename, users=False):
            if not isinstance(database_keys, str) and len(database_keys) > 1:
                data = self.generate_dataframe(
                    database_keys, header.strip().split(",")[1:]
                )
            else:
                key = (
                    database_keys
                    if isinstance(database_keys, str)
                    else database_keys[0]
                )
                data = self.generate_series(key)
            if users:
                data = await self.convert_users(data)
            with StringIO() as f:
                f.write(header)
                data.to_csv(f, header=False)
                with BytesIO(f.getvalue().encode("utf-8")) as b:
                    files.append(discord.File(b, filename))

        if ctx.interaction is not None:
            await ctx.typing()

        logger.info("exporting freq command")
        await _export_helper(
            "frequency.command:global",
            "command,amount used\n",
            "command_frequency.csv",
            users=False,
        )

        logger.info("exporting freq item")
        await _export_helper(
            "frequency.item:global",
            f"{config.options['id_type']},amounts seen\n",
            f"{config.options['id_type']}_frequency.csv",
            users=False,
        )

        logger.info("exporting streaks")
        await _export_helper(
            ["streak:global", "streak.max:global"],
            "username#discrim,current streak,max streak\n",
            "streaks.csv",
            True,
        )

        logger.info("exporting missed")
        keys = tuple(
            sorted(
                map(
                    lambda x: x.decode("utf-8"),
                    database.scan_iter(match="daily.incorrect:????-??-??", count=5000),
                )
            )
        )
        titles = ",".join(map(lambda x: x.split(":")[1], keys))
        keys = ("incorrect:global",) + keys
        await _export_helper(
            keys,
            f"{config.options['id_type']},total missed,{titles}\n",
            "missed.csv",
            users=False,
        )

        logger.info("exporting scores")
        keys = tuple(
            sorted(
                map(
                    lambda x: x.decode("utf-8"),
                    database.scan_iter(match="daily.score:????-??-??", count=5000),
                )
            )
        )
        titles = ",".join(map(lambda x: x.split(":")[1], keys))
        keys = ("users:global",) + keys
        await _export_helper(
            keys, f"username#discrim,total score,{titles}\n", "scores.csv", users=True
        )

        await ctx.send(files=files)


async def setup(bot):
    await bot.add_cog(Stats(bot))



# blueprints/verify.py | Flask routes relating to image verification
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

from flask import Blueprint, abort, jsonify, request, send_file, url_for

import sciolyid.config as config
import sciolyid.web.functions.webhooks as webhooks
from sciolyid.web.config import logger
from sciolyid.web.functions.images import filename_lookup, find_duplicates
from sciolyid.web.functions.user import get_user_id
from sciolyid.web.git import verify_repo
from sciolyid.web.tasks import database

bp = Blueprint("verify", __name__, url_prefix="/verify")


@bp.route("/", methods=("GET",))
def verify_files():
    logger.info("endpoint: verify")
    user_id: str = get_user_id()
    verify_repo.remote("origin").pull()

    lookup = filename_lookup(
        os.path.abspath(
            config.options["validation_local_dir"]
            + config.options["validation_repo_dir"]
        )
    )
    ids: tuple = tuple(lookup.keys())

    if len(ids) == 0:
        return jsonify({"end": True})

    offset: int = request.args.get("offset", 0, int)

    seen: set = set(
        map(lambda x: x.decode(), database.smembers(f"sciolyid.verify.user:{user_id}"))
    )
    image_id: str = ""
    for i in range(len(ids)):
        offset = (i + offset) % len(ids)
        id_ = ids[offset]
        if id_ not in seen:
            image_id = id_
            break
        if i == len(ids) - 1:
            return jsonify({"end": True})

    output = {}
    output["offset"] = offset
    output["url"] = url_for(".send_image", image_id=image_id)
    output["duplicates"] = find_duplicates(lookup[image_id], ignore_verify=True)
    output["item"] = lookup[image_id].split("/")[-2]
    output["id"] = image_id
    return jsonify(output)


@bp.route("/image/<string:image_id>", methods=("GET",))
def send_image(image_id: str):
    logger.info("endpoint: verify.send_image")
    get_user_id()

    lookup = filename_lookup(
        os.path.abspath(
            config.options["validation_local_dir"]
            + config.options["validation_repo_dir"]
        )
    )
    image_path = lookup.get(image_id, None)
    if not image_path:
        abort(404, "Filename not found!")
    return send_file(image_path)


@bp.route("/confirm", methods=("POST",))
def confirm():
    logger.info("endpoint: verify.confirm")
    user_id: str = get_user_id()

    confirmation = request.form["confirmation"]
    if confirmation not in ("valid", "invalid", "duplicate"):
        abort(400, "invalid confirmation field")

    lookup = filename_lookup(
        os.path.abspath(
            config.options["validation_local_dir"]
            + config.options["validation_repo_dir"]
        )
    )
    image_id = request.form["id"]
    if image_id not in lookup.keys():
        abort(400, "invalid id")

    if database.sismember(f"sciolyid.verify.user:{user_id}", image_id):
        abort(400, "You've already confirmed this image!")

    database.zincrby(f"sciolyid.verify.images:{confirmation}", 1, image_id)
    database.sadd(f"sciolyid.verify.user:{user_id}", image_id)
    webhooks.send("verify", user_id=user_id, action=confirmation)
    return jsonify({"success": True})


@bp.route("/stats", methods=("GET",))
def stats():
    logger.info("endpoint: verify.stats")
    get_user_id()

    image_id = request.args.get("id", "", str)
    lookup = filename_lookup(
        os.path.abspath(
            config.options["validation_local_dir"]
            + config.options["validation_repo_dir"]
        )
    )
    if image_id not in lookup.keys():
        abort(400, "invalid id")

    output = dict()
    for confirmation in ("valid", "duplicate", "invalid"):
        output[confirmation] = int(
            database.zscore(f"sciolyid.verify.images:{confirmation}", image_id) or 0
        )

    return jsonify(output)
